from pyneuphonic.agents import Agent
from pyneuphonic.client import Neuphonic
from pyneuphonic.models import TTSConfig, WebsocketEvents, AgentConfig
from pyneuphonic.player import AudioPlayer, AsyncAudioPlayer, AsyncAudioRecorder
from pyneuphonic._utils import save_audio, async_save_audio



from pydantic import BaseModel as BaseModel, field_validator, ConfigDict, Field
from typing import List, Optional, Callable, Awaitable, Union, TypedDict
import base64
from enum import Enum
from typing import Generic, TypeVar


def to_dict(model: BaseModel):
    """Returns a pydantic model as a dictionary, excluding items with None values."""
    return {k: v for k, v in model.model_dump().items() if v is not None}


T = TypeVar('T')


class BaseConfig(BaseModel):
    model_config = ConfigDict(extra='allow')

    def to_query_params(self) -> str:
        """Generate a query parameter string from the object, excluding None values."""
        params = to_dict(self)
        return '&'.join(f'{key}={value}' for key, value in params.items())


class AgentConfig(BaseConfig):
    """
    API parameters for the /agent endpoint.
    """

    agent_id: Optional[str] = Field(
        default=None,
        description='ID of the selected agent. If None, a default agent will be used.',
        examples=['da78ea32-9225-436e-b10d-d5b101bb01a6'],  # example agent_id
    )

    tts_model: Optional[str] = Field(
        default=None,
        description='The Neuphonic model to be used for text-to-speech synthesis.',
        examples=['neu_fast', 'neu_hq'],
    )

    endpointing: Optional[float] = Field(
        default=None,
        description=(
            'Duration (in milliseconds) the speech recognition program will listen for '
            'silence in the received audio before concluding the user has finished speaking.'
        ),
        examples=[50, 100, 1000],
    )

    mode: Optional[str] = Field(
        default='asr-llm-tts',
        description=(
            'Mode of agent usage. `asr-llm-tts` indicates audio input and output. '
            '`llm-tts` indicates text input and audio output.'
        ),
        examples=['asr-llm-tts', 'llm-tts'],
    )

    incoming_sampling_rate: Optional[int] = Field(
        default=16000,
        description=(
            'Sampling rate of the audio sent to the server. Lower rates generally '
            'yield faster transcription.'
        ),
        examples=[8000, 16000, 22050],
    )

    return_sampling_rate: Optional[int] = Field(
        default=22050,
        description='Sampling rate of the audio returned from the server.',
        examples=[8000, 16000, 22050],
    )

    incoming_encoding: Optional[str] = Field(
        default='pcm_linear',
        description='Encoding of the audio sent to the server.',
        examples=['pcm_linear', 'pcm_mulaw'],
    )

    return_encoding: Optional[str] = Field(
        default='pcm_linear',
        description='Encoding of the audio returned from the server.',
        examples=['pcm_linear', 'pcm_mulaw'],
    )


class TTSConfig(BaseConfig):
    """
    Model parameters for the text-to-speech endpoints.
    """

    speed: Optional[float] = Field(
        default=1.0,
        description='Playback speed of the audio.',
        examples=[0.7, 1.0, 1.5],
    )

    temperature: Optional[float] = Field(
        default=None,
        description='Randomness introduced into the text-to-speech model. Range: 0 to 1.0.',
        examples=[0.5, 0.7],
    )

    model: Optional[str] = Field(
        default=None,
        description='The text-to-speech model to use.',
        examples=['neu_fast', 'neu_hq'],
    )

    lang_code: str = Field(
        default='en',
        description='Language code for the desired language.',
        examples=['en'],
    )

    voice_id: Optional[str] = Field(
        default=None,
        description=(
            'The voice ID for the desired voice. Ensure this voice ID is available for the '
            'selected model.'
        ),
        examples=['8e9c4bc8-3979-48ab-8626-df53befc2090'],
    )

    sampling_rate: Optional[int] = Field(
        default=22050,
        description='Sampling rate of the audio returned from the server.',
        examples=[8000, 16000, 22050],
    )

    encoding: Optional[str] = Field(
        default='pcm_linear',
        description='Encoding of the audio returned from the server.',
        examples=['pcm_linear', 'pcm_mulaw'],
    )


class APIResponse(BaseModel, Generic[T]):
    """All API responses will be typed with this pydantic model."""

    model_config = ConfigDict(extra='allow')

    data: Optional[T] = Field(
        default=None,
        description='API response data. Contains data on a successful response.',
    )

    metadata: Optional[dict] = Field(
        default=None,
        description=(
            'Additional metadata from the API. Includes pagination metadata for paginated '
            'endpoints.'
        ),
    )

    """
    The following fields are only for responses from SSE endpoints.
    """
    status_code: Optional[int] = Field(
        default=None,
        description=(
            'Status code of the API response. Only set for responses on SSE endpoints.'
        ),
        examples=[200, 400],
    )

    errors: Optional[List[str]] = Field(
        default=None,
        description=(
            'All errors associated with the SSE response, if the status_code is non-2XX.'
        ),
    )


class VoiceObject(TypedDict):
    """TypedDict representing a voice object with its attributes."""

    voice_id: str
    """
    The voice ID.
    Examples: ['8e9c4bc8-3979-48ab-8626-df53befc2090']
    """

    name: str
    """
    The name of the voice.
    Examples: ['Holly', 'Annie', 'Miles']
    """

    tags: List[str]
    """
    A list of tags describing the voice.
    Examples: [['Male', 'American', 'Forties', 'Narrator'], ['Female', 'British', 'Twenties', 'Excited']]
    """

    model_availability: List[str]
    """
    A list of models that this voice is available on.
    Examples: [['neu_fast', 'neu_hq'], ['neu_hq']]
    """

    type: str
    """
    Indicates whether this is a standard voice provided by Neuphonic, or a cloned voice.
    """

    lang_code: str
    """
    The language code associated with this voice.
    """


class AgentObject(TypedDict):
    """TypedDict representing an agent object with its attributes."""

    id: str
    """
    The agent ID.
    Examples: ['1234abcd-5678-efgh-9101-ijklmnopqrst']
    """

    name: str
    """
    The name of the agent.
    Examples: ['Helpful Receptionist', 'Delivery Person', 'John Smith']
    """

    prompt: Optional[str]
    """
    The prompt used by the agent.
    Examples: ['You are a helpful agent. Answer in 20 words or less.']
    """

    greeting: Optional[str]
    """
    The greeting message of the agent.
    Examples: ['Hello! How can I help you?', 'Hi there! What can I do for you today?']
    """


class AudioBaseModel(BaseModel):
    """
    Base model for any models containing audio.
    """

    model_config = ConfigDict(extra='allow')

    audio: Optional[bytes] = Field(
        default=None,
        description=(
            'Audio received from the server. The server returns audio as a base64 encoded string, '
            'which will be parsed into bytes by the field_validator.'
        ),
    )

    @field_validator('audio', mode='before')
    def validate(cls, v: Optional[Union[str, bytes]]) -> Optional[bytes]:
        """Convert the received audio from the server from base64 into bytes."""
        if isinstance(v, str):
            return base64.b64decode(v)
        elif isinstance(v, bytes):
            return v
        elif v is None:
            return None

        raise ValueError('`audio` must be a base64 encoded string or bytes.')


class TTSResponse(AudioBaseModel):
    """Structure of data received from TTS endpoints, when using any client in `Neuphonic.tts`."""

    text: Optional[str] = Field(
        default=None, description='Text corresponding to the audio snippet.'
    )

    sampling_rate: Optional[int] = Field(
        default=None,
        description='Sampling rate of the audio snippet.',
        examples=[8000, 16000, 22050],
    )


class AgentResponse(AudioBaseModel):
    """Structure of data received from Agent endpoints, when using any client in `Neuphonic.agent`."""

    type: str = Field(
        description='Type of message being sent by the server',
        examples=['llm_response', 'audio_response', 'user_transcript'],
    )

    text: Optional[str] = Field(
        default=None,
        description=(
            'Corresponding text if the `type` is `llm_response` or `user_transcript`.'
        ),
    )


class WebsocketEvents(Enum):
    """Enum describing all valid websocket events that callbacks can be bound to."""

    OPEN: str = 'open'
    MESSAGE: str = 'message'
    CLOSE: str = 'close'
    ERROR: str = 'error'


class WebsocketEventHandlers(BaseModel):
    """Pydantic model to hold all websocket callbacks."""

    open: Optional[Callable[[], Awaitable[None]]] = None
    message: Optional[Callable[[APIResponse[T]], Awaitable[None]]] = None
    close: Optional[Callable[[], Awaitable[None]]] = None
    error: Optional[Callable[[Exception], Awaitable[None]]] = None



import asyncio
import websockets
from typing import Callable, Union
import json
from abc import ABC, abstractmethod

from pyneuphonic._endpoint import Endpoint
from pyneuphonic.models import (
    WebsocketEventHandlers,
    TTSConfig,
    APIResponse,
    TTSResponse,
    WebsocketEvents,
    BaseConfig,
    AgentConfig,
    AgentResponse,
)
from pydantic import BaseModel


class AsyncWebsocketBase(Endpoint, ABC):
    """
    Abstract base class for asynchronous websocket clients.

    Parameters
    ----------
    api_key : str
        The API key for authentication.
    base_url : str
        The base URL for the websocket connection.
    response_type : BaseModel
        The type of response expected from the websocket. This will be one of TTSResponse and
        AgentResponse.
    """

    def __init__(
        self,
        api_key: str,
        base_url: str,
        response_type: BaseModel,
    ):
        super().__init__(api_key=api_key, base_url=base_url)

        self.event_handlers = WebsocketEventHandlers()
        self.message_queue = asyncio.Queue()

        self._ws = None
        self._tasks = []

        self.response_type = response_type

    @abstractmethod
    def url(self, config: Union[BaseConfig, dict]) -> str:
        """
        Construct the URL for the websocket connection.

        Parameters
        ----------
        config : Union[BaseConfig, dict]
            Configuration for the websocket connection. This is required to extract the query
            parameters.

        Returns
        -------
        str
            The constructed URL. E.g.: wss://eu-west-1.api.neuphonic.com/speak/en?model=neu_hq
        """
        pass

    def on(self, event: WebsocketEvents, handler: Callable):
        """
        Register an event handler for a specific websocket event.

        Parameters
        ----------
        event : WebsocketEvents
            The event to handle.
        handler : Callable
            The function to call when the event occurs.

        Raises
        ------
        ValueError
            If the event is not a valid WebsocketEvents.
        """
        if event not in WebsocketEvents:
            raise ValueError(f'Event "{event}" is not a valid event.')

        setattr(self.event_handlers, event.value, handler)

    @abstractmethod
    async def open(self, config: Union[BaseConfig, dict]):
        """
        Open the websocket connection. After this function is called, the websocket will start
        listening and sending requests.

        Parameters
        ----------
        config : Union[BaseConfig, dict]
            Configuration for the websocket connection.
        """
        self._ws = await websockets.connect(
            self.url(config),
            ssl=self.ssl_context,
            extra_headers=self.headers,
        )

        if self.event_handlers.open is not None:
            await self.event_handlers.open()

        receive_task = asyncio.create_task(self._receive())
        self._tasks.append(receive_task)

    async def _receive(self):
        """
        Receive messages from the websocket and handle them. This is created and launched as an
        async task when when self.open is called.
        """
        try:
            async for message in self._ws:
                if isinstance(message, str):
                    message = APIResponse[self.response_type](**json.loads(message))

                    if self.event_handlers.message is not None:
                        await self.event_handlers.message(message)
                    else:
                        await self.message_queue.put(message)
        except Exception as e:
            if self.event_handlers.error is not None:
                await self.event_handlers.error(e)
        finally:
            if self.event_handlers.close:
                await self.event_handlers.close()

            await self.close()

    async def send(self, message: Union[str, dict], *args, **kwargs):
        """
        Send a message through the websocket.

        Parameters
        ----------
        message : Union[str, dict]
            The message to send. Must be a string or a dictionary.

        Raises
        ------
        AssertionError
            If the message is not a string or dictionary.
        """
        assert isinstance(
            message, (str, dict)
        ), 'Message must be an instance of str or dict'

        message = message if isinstance(message, str) else json.dumps(message)

        await self._ws.send(message)

    async def receive(self):
        """
        Receive a message from the message queue. This is where messages are placed if no message
        handler has been set.

        Returns
        -------
        APIResponse
            The next message in the queue.
        """
        return await self.message_queue.get()

    async def close(self):
        """
        Close the websocket connection and cancel all tasks.
        """
        for task in self._tasks:
            task.cancel()

            try:
                await task
            except asyncio.CancelledError:
                pass

        await self._ws.close()


class AsyncTTSWebsocketClient(AsyncWebsocketBase):
    """
    Asynchronous websocket client for Text-to-Speech (TTS) operations.

    Parameters
    ----------
    api_key : str
        The API key for authentication.
    base_url : str
        The base URL for the websocket connection.
    """

    def __init__(self, api_key: str, base_url: str):
        super().__init__(
            api_key=api_key,
            base_url=base_url,
            response_type=TTSResponse,
        )

    def url(self, config: Union[TTSConfig, dict]) -> str:
        """
        See AsyncWebsocketClientBase.url
        """
        if not isinstance(config, TTSConfig):
            config = TTSConfig(**config)

        return f'{self.ws_url}/speak/{config.lang_code}?{config.to_query_params()}'

    async def open(self, tts_config: Union[TTSConfig, dict] = TTSConfig()):
        """
        See AsyncWebsocketClientBase.open
        """
        await super().open(tts_config)

    async def send(self, message: Union[str, dict], autocomplete=False):
        """
        Send a message through the TTS websocket. This handles autocompletion of messages as well.

        Parameters
        ----------
        message : Union[str, dict]
            The message to send.
        autocomplete : bool, optional
            Whether to send an autocomplete '<STOP>' signal, by default False
        """
        await super().send(message=message)

        if autocomplete:
            await self._ws.send(json.dumps({'text': ' <STOP>'}))

    async def complete(self):
        """
        Send a completion signal '<STOP>' through the TTS websocket.
        """
        await self.send({'text': ' <STOP>'}, autocomplete=False)


class AsyncAgentWebsocketClient(AsyncWebsocketBase):
    """
    Asynchronous websocket client for /agents operations.

    Parameters
    ----------
    api_key : str
        The API key for authentication.
    base_url : str
        The base URL for the websocket connection.
    """

    def __init__(self, api_key: str, base_url: str):
        super().__init__(
            api_key=api_key,
            base_url=base_url,
            response_type=AgentResponse,
        )

    def url(self, config: Union[AgentConfig, dict]) -> str:
        """
        See AsyncWebsocketClientBase.url
        """
        if not isinstance(config, AgentConfig):
            config = AgentConfig(**config)

        return f'{self.ws_url}/agents?{config.to_query_params()}'

    async def open(self, agent_config: Union[TTSConfig, dict] = AgentConfig()):
        """
        See AsyncWebsocketClientBase.open
        """
        await super().open(agent_config)



import wave

from typing import Optional, Iterator, Union, AsyncIterator

from pyneuphonic.models import APIResponse, TTSResponse

def save_audio(
    audio_bytes: Union[bytes, bytearray, Iterator[APIResponse[TTSResponse]]],
    file_path: str,
    sample_rate: Optional[int] = 22050,
):
    """
    Takes in an audio buffer and saves it to a .wav file.

    Parameters
    ----------
    audio_bytes
        The audio buffer to save. This is all the bytes returned from the server.
    file_path
        The file path you want to save the audio to.
    sample_rate
        The sample rate of the audio you want to save. Default is 22050.
    """
    if isinstance(audio_bytes, bytes) or isinstance(audio_bytes, bytearray):
        with wave.open(file_path, 'wb') as wav_file:
            wav_file.setnchannels(1)
            wav_file.setsampwidth(2)
            wav_file.setframerate(sample_rate)
            wav_file.writeframes(bytes(audio_bytes))
    elif isinstance(audio_bytes, Iterator):
        with wave.open(file_path, 'wb') as wav_file:
            wav_file.setnchannels(1)
            wav_file.setsampwidth(2)
            wav_file.setframerate(sample_rate)
            for message in audio_bytes:
                if not isinstance(message, APIResponse[TTSResponse]):
                    raise ValueError(
                        '`audio_bytes` must be an Iterator yielding an object of type'
                        '`pyneuphonic.models.APIResponse[TTSResponse]`'
                    )
                wav_file.writeframes(message.data.audio)


async def async_save_audio(
    audio_bytes: Union[bytes, bytearray, AsyncIterator[APIResponse[TTSResponse]]],
    file_path: str,
    sample_rate: Optional[int] = 22050,
):
    """
    Takes in an audio buffer and saves it to a .wav file.

    Parameters
    ----------
    audio_bytes
        The audio buffer to save. This is all the bytes returned from the server.
    file_path
        The file path you want to save the audio to.
    sample_rate
        The sample rate of the audio you want to save. Default is 22050.
    """
    if isinstance(audio_bytes, bytes) or isinstance(audio_bytes, bytearray):
        with wave.open(file_path, 'wb') as wav_file:
            wav_file.setnchannels(1)
            wav_file.setsampwidth(2)
            wav_file.setframerate(sample_rate)
            wav_file.writeframes(bytes(audio_bytes))
    elif isinstance(audio_bytes, AsyncIterator):
        with wave.open(file_path, 'wb') as wav_file:
            wav_file.setnchannels(1)
            wav_file.setsampwidth(2)
            wav_file.setframerate(sample_rate)
            async for message in audio_bytes:
                if not isinstance(message, APIResponse[TTSResponse]):
                    raise ValueError(
                        '`audio_bytes` must be an AsyncIterator yielding an object of type'
                        '`pyneuphonic.models.APIResponse[TTSResponse]`'
                    )
                wav_file.writeframes(message.data.audio)



import asyncio
import logging

from typing import Union, Optional, Iterator, AsyncIterator
from pyneuphonic.models import APIResponse, TTSResponse
from pyneuphonic._utils import save_audio
from base64 import b64encode
import time

try:
    import pyaudio
except ModuleNotFoundError:
    logging.warning(
        '(pyneuphonic) `pyaudio` is not installed, so audio playback and audio recording'
        ' functionality will not be enabled, and attempting to use this functionality may'
        ' throw errors. `pip install pyaudio` to resolve. This message may be ignored if'
        ' audio playback and recording features are not required.'
    )


class AudioPlayer:
    """Handles audio playback and audio exporting."""

    def __init__(self, sampling_rate: int = 22050):
        """
        Initialize with a default sampling rate.

        Parameters
        ----------
        sampling_rate : int
            The sample rate for audio playback.
        """
        self.sampling_rate = sampling_rate
        self.audio_player = None
        self.stream = None
        self.audio_bytes = bytearray()

        # indicates when audio will stop playing
        self._playback_end = time.perf_counter()

    @property
    def is_playing(self):
        """Returns True if there is audio currently playing."""
        return time.perf_counter() < self._playback_end

    def open(self):
        """Open the audio stream for playback. `pyaudio` must be installed."""
        self.audio_player = pyaudio.PyAudio()  # create the PyAudio player

        # start the audio stream, which will play audio as and when required
        self.stream = self.audio_player.open(
            format=pyaudio.paInt16, channels=1, rate=self.sampling_rate, output=True
        )

    def play(self, data: Union[bytes, Iterator[APIResponse[TTSResponse]]]):
        """
        Play audio data or automatically stream over SSE responses and play the audio.

        Parameters
        ----------
        data : Union[bytes, Iterator[TTSResponse]]
            The audio data to play, either as bytes or an iterator of TTSResponse.
        """
        if isinstance(data, bytes):
            if self.stream:
                duration = len(data) / (2 * self.sampling_rate)

                if self.is_playing:
                    self._playback_end += duration
                else:
                    self._playback_end = time.perf_counter() + duration

                self.stream.write(data)
            self.audio_bytes += data
        elif isinstance(data, Iterator):
            for message in data:
                if not isinstance(message, APIResponse[TTSResponse]):
                    raise ValueError(
                        '`data` must be an Iterator yielding an object of type'
                        '`pyneuphonic.models.APIResponse[TTSResponse]`'
                    )

                self.play(message.data.audio)
        else:
            raise TypeError(
                '`data` must be of type bytes or an Iterator of APIResponse[TTSResponse]'
            )

    def close(self):
        """Close the audio stream and terminate resources."""
        if self.stream:
            self.stream.stop_stream()
            self.stream.close()
            self.stream = None
        if self.audio_player:
            self.audio_player.terminate()
            self.audio_player = None

    def save_audio(
        self,
        file_path: str,
        sample_rate: Optional[int] = 22050,
    ):
        """Saves the audio using pynuephonic.save_audio"""
        save_audio(
            audio_bytes=self.audio_bytes, sample_rate=sample_rate, file_path=file_path
        )

    def __enter__(self):
        """Enter the runtime context related to this object."""
        self.open()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit the runtime context related to this object."""
        self.close()


class AsyncAudioPlayer(AudioPlayer):
    def __init__(self, sampling_rate: int = 22050):
        super().__init__(sampling_rate)

    async def open(self):
        super().open()

    async def play(self, data: Union[bytes, AsyncIterator[APIResponse[TTSResponse]]]):
        if isinstance(data, bytes):
            await asyncio.to_thread(super().play, data)
        elif isinstance(data, AsyncIterator):
            async for message in data:
                if not isinstance(message, APIResponse[TTSResponse]):
                    raise ValueError(
                        '`data` must be an AsyncIterator yielding an object of type'
                        '`pyneuphonic.models.APIResponse[TTSResponse]`'
                    )

                await self.play(message.data.audio)
        else:
            raise TypeError(
                '`data` must be of type bytes or an AsyncIterator of APIResponse[TTSResponse]'
            )

    async def close(self):
        super().close()

    async def __aenter__(self):
        """Enter the runtime context related to this object."""
        await self.open()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        """Exit the runtime context related to this object."""
        await self.close()


class AsyncAudioRecorder:
    def __init__(
        self, sampling_rate: int = 16000, websocket=None, player: AudioPlayer = None
    ):
        self.p = None
        self.stream = None
        self.sampling_rate = sampling_rate

        self._ws = websocket
        self.player = player
        self._queue = asyncio.Queue()  # Use a queue to handle audio data asynchronously

        self._tasks = []

    async def _send(self):
        while True:
            try:
                # Wait for audio data from the queue
                data = await self._queue.get()

                if self.player is not None and not self.player.is_playing:
                    await self._ws.send({'audio': b64encode(data).decode('utf-8')})
            except Exception as e:
                logging.error(f'Error in _send: {e}')

    def _callback(self, in_data, frame_count, time_info, status):
        try:
            # Enqueue the incoming audio data for processing in the async loop
            self._queue.put_nowait(in_data)
        except asyncio.QueueFull:
            logging.error('Audio queue is full! Dropping frames.')
        return None, pyaudio.paContinue

    async def record(self):
        self.p = pyaudio.PyAudio()

        self.stream = self.p.open(
            format=pyaudio.paInt16,
            channels=1,
            rate=self.sampling_rate,
            input=True,
            stream_callback=self._callback,  # Use the callback function
        )

        self.stream.start_stream()  # Explicitly start the stream

        if self._ws is not None:
            send_task = asyncio.create_task(self._send())
            self._tasks.append(send_task)

    async def close(self):
        if self.stream:
            self.stream.stop_stream()
            self.stream.close()
            self.stream = None
        if self.p:
            self.p.terminate()
            self.p = None

        for task in self._tasks:
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                pass

    async def __aenter__(self):
        await self.record()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.close()



import asyncio
import aioconsole

from pyneuphonic.client import Neuphonic
from pyneuphonic.models import APIResponse, AgentResponse, AgentConfig, WebsocketEvents
from pyneuphonic.player import AsyncAudioPlayer, AsyncAudioRecorder


def default_on_message(message: APIResponse[AgentResponse]):
    """
    Default callback function to handle messages from the server.

    Parameters
    ----------
    message : APIResponse[AgentResponse]
        The message received from the server, containing the type and text.
    """
    if message.data.type == 'user_transcript':
        print(f'User: {message.data.text}')
    elif message.data.type == 'llm_response':
        print(f'Agent: {message.data.text}')


class Agent:
    def __init__(
        self, client: Neuphonic, mute=False, on_message=default_on_message, **kwargs
    ):
        """
        Initialize an Agent instance.

        Parameters
        ----------
        client : Neuphonic
            The Neuphonic client instance.
        mute : bool, optional
            If True, the agent will not play audio responses. Default is False.
        on_message : callable, optional
            A callback function to handle messages from the server. Default is default_on_message.
        **kwargs
            Additional keyword arguments to configure the agent. See the `AgentConfig` model for a
            full list of agent configuration parameters.
        """
        self.config = AgentConfig(**kwargs)
        self.mute = mute

        self.ws = client.agents.AsyncWebsocketClient()

        self.player = None
        if not self.mute:
            self.player = AsyncAudioPlayer()

        if 'asr' in self.config.mode:
            # passing in the websocket object will automatically forward audio to the server
            self.recorder = AsyncAudioRecorder(
                sampling_rate=self.config.incoming_sampling_rate,
                websocket=self.ws,
                player=self.player,
            )

        self.on_message_hook = on_message
        self._tasks = []

    async def on_message(self, message: APIResponse[AgentResponse]):
        """
        Handle incoming messages from the server.

        Parameters
        ----------
        message : APIResponse[AgentResponse]
            The message received from the server, containing the type and content.
        """
        # server will return 3 types of messages: audio_response, user_transcript, llm_response
        if message.data.type == 'audio_response':
            if not self.mute:
                await self.player.play(message.data.audio)

        if self.on_message_hook is not None and callable(self.on_message_hook):
            self.on_message_hook(message)

    async def start(self):
        """
        Start the agent, opening necessary connections and handling user input.
        """
        self.ws.on(WebsocketEvents.MESSAGE, self.on_message)
        self.ws.on(WebsocketEvents.CLOSE, self.on_close)

        async def run_agent():
            if not self.mute:
                await self.player.open()
            await self.ws.open(self.config)

            if 'asr' in self.config.mode:
                await self.recorder.record()

                try:
                    while True:
                        await asyncio.sleep(0.01)
                except KeyboardInterrupt:
                    await self.close()

            else:
                while True:
                    user_text = await aioconsole.ainput(
                        "\nEnter text to speak (or 'quit' to exit): "
                    )

                    if user_text.lower() == 'quit':
                        break

                    await self.ws.send({'text': user_text})
                    await asyncio.sleep(1)  # simply for formatting

            await self.close()

        run_task = asyncio.create_task(run_agent())
        self._tasks.append(run_task)

    async def on_close(self):
        """
        Handle the closing of connections and cleanup resources.
        """
        if not self.mute:
            await self.player.close()
        if 'asr' in self.config.mode:
            await self.recorder.close()

    async def stop(self):
        """
        Close the agent, including websocket and any active resources.
        """
        for task in self._tasks:
            task.cancel()

            try:
                await task
            except asyncio.CancelledError:
                pass

        await self.ws.close()
        await self.on_close()



from typing import Optional
import httpx

from pyneuphonic._endpoint import Endpoint
from pyneuphonic._websocket import AsyncAgentWebsocketClient
from pyneuphonic.models import APIResponse, AgentObject  # noqa: F401


class Agents(Endpoint):
    def list(
        self,
    ) -> APIResponse[dict]:
        """
        List created agents.

        By default this endpoint returns only `id` and `name` for every agent, provide the `agent_id`
        parameter to get all the fields for a specific agent.

        Parameters
        ----------
        agent_id
            The ID of the agent to fetch. If None, fetches all agents.

        Returns
        -------
        APIResponse[dict]
            response.data['agent'] will be an object of type AgentObject.

        Raises
        ------
        httpx.HTTPStatusError
            If the request fails to fetch.
        """
        response = httpx.get(
            f'{self.http_url}/agents',
            headers=self.headers,
            timeout=self.timeout,
        )

        self.raise_for_status(response=response, message='Failed to fetch agents.')

        return APIResponse(**response.json())

    def get(
        self,
        agent_id: str,
    ) -> APIResponse[dict]:
        """
        List created agents.

        By default this endpoint returns only `id` and `name` for every agent, provide the `agent_id`
        parameter to get all the fields for a specific agent.

        Parameters
        ----------
        agent_id
            The ID of the agent to fetch. If None, fetches all agents.

        Returns
        -------
        APIResponse[dict]
            response.data['agent'] will be an object of type AgentObject.

        Raises
        ------
        httpx.HTTPStatusError
            If the request fails to fetch.
        """
        response = httpx.get(
            f'{self.http_url}/agents/{agent_id}',
            headers=self.headers,
            timeout=self.timeout,
        )

        self.raise_for_status(response=response, message='Failed to fetch agents.')

        return APIResponse(**response.json())

    def create(
        self,
        name: str,
        prompt: Optional[str] = None,
        greeting: Optional[str] = None,
    ) -> APIResponse[dict]:
        """
        Create a new agent.

        Parameters
        ----------
        name
            The name of the agent.
        prompt
            The prompt for the agent.
        greeting
            The initial greeting message for the agent.

        Returns
        -------
        APIResponse[dict]
            response.data will contain a success message on successful creation.

        Raises
        ------
        httpx.HTTPStatusError
            If the request fails to create.
        """
        data = {
            'name': name,
            'prompt': prompt,
            'greeting': greeting,
        }

        response = httpx.post(
            f'{self.http_url}/agents',
            json=data,
            headers=self.headers,
            timeout=self.timeout,
        )

        self.raise_for_status(response=response, message='Failed to create agent.')

        return APIResponse(**response.json())

    def delete(
        self,
        agent_id: str,
    ) -> APIResponse[dict]:
        """
        Delete an agent.

        Parameters
        ----------
        agent_id : str
            The ID of the agent to delete.

        Returns
        -------
        APIResponse[dict]
            response.data will contain a delete message on successful deletion.

        Raises
        ------
        httpx.HTTPStatusError
            If the request fails to delete.
        """
        response = httpx.delete(
            f'{self.http_url}/agents/{agent_id}',
            headers=self.headers,
            timeout=self.timeout,
        )

        self.raise_for_status(response=response, message='Failed to delete agent.')

        return APIResponse(**response.json())

    def AsyncWebsocketClient(self):
        return AsyncAgentWebsocketClient(api_key=self._api_key, base_url=self._base_url)



from typing import List, Optional

import httpx

from ._endpoint import Endpoint
from .models import APIResponse, VoiceObject  # noqa: F401


class Voices(Endpoint):
    def list(self) -> APIResponse[dict]:
        """Lists all voices in your voice library.

        Returns
        -------
        APIResponse[dict]
            response.data['voices'] will be a list of VoiceObject objects.
        """
        response = httpx.get(
            f'{self.http_url}/voices',
            headers=self.headers,
            timeout=self.timeout,
        )

        if not response.is_success:
            raise httpx.HTTPStatusError(
                f'Failed to fetch voices. Status code: {response.status_code}. Error: {response.text}',
                request=response.request,
                response=response,
            )

        return APIResponse(**response.json())

    def _get_voice_id_from_name(self, voice_name) -> str:
        """Gets the voice_id given a voice name.

        Parameters
        ----------
        voice_name : str
            The name of the voice to retrieve the voice_id for.

        Raises
        ------
        ValueError
            Raised if there is no voice with the provided name.
        """
        response = self.list()
        voices = response.data['voices']

        try:
            # extract the voice_id for the requested voice from the list of all voices
            return next(
                voice['voice_id'] for voice in voices if voice['name'] == voice_name
            )
        except StopIteration as e:
            raise ValueError(f'No voice found with the name {voice_name}.')

    def get(self, voice_id: str = None, voice_name: str = None) -> APIResponse[dict]:
        """Get information about specific voice.

        Parameters
        ----------
        voice_name : str
            The voice name you want to retreive the information for.
        voice_id : str
            The voice id you want to retreive the information for.

        Returns
        -------
        APIResponse[dict]
            response.data['voice'] will be a single VoiceObject object.

        Raises
        ------
        httpx.HTTPStatusError
            If the request to clone the voice fails.
        """

        # Accept case if user only provide name
        if voice_id is None and voice_name is None:
            raise ValueError('Please provide one of voice_id or voice_name')
        if not voice_id:
            voice_id = self._get_voice_id_from_name(voice_name=voice_name)

        response = httpx.get(
            f'{self.http_url}/voices/{voice_id}',
            headers=self.headers,
            timeout=self.timeout,
        )

        if not response.is_success:
            raise httpx.HTTPStatusError(
                f'Failed to fetch voice. Status code: {response.status_code}. Error: {response.text}',
                request=response.request,
                response=response,
            )

        return APIResponse(**response.json())

    def clone(
        self, voice_name: str, voice_file_path: str, voice_tags: List[str] = []
    ) -> APIResponse[dict]:
        """
        Clone a voice by uploading a file with the specified name and tags.

        Parameters
        ----------
        voice_name : str
            The name of the new cloned voice.
        voice_file_path : str
            Path to the voice file (e.g., a .wav file) to be uploaded.
        voice_tags : List[str]
            Tags associated with the voice. Default is an empty list.

        Returns
        -------
        APIResponse[dict]
            response.data will contain a success message with voice_id of the cloned voice.

        Raises
        ------
        httpx.HTTPStatusError
            If the request to clone the voice fails.
        """

        # Convert voice tags to a string
        if voice_tags:
            voice_tags = (', ').join(voice_tags)

        # Prepare the multipart form-data payload
        params = {
            'voice_tags': voice_tags,
        }
        files = {'voice_file': open(voice_file_path, 'rb')}

        # Send the POST request with voice_name as a query parameter
        response = httpx.post(
            f'{self.http_url}/voices?voice_name={voice_name}',
            params=params,
            files=files,
            headers=self.headers,
            timeout=self.timeout,
        )

        # Handle response errors
        if not response.is_success:
            raise httpx.HTTPStatusError(
                f'Failed to clone voice. Status code: {response.status_code}. Error: {response.text}',
                request=response.request,
                response=response,
            )

        # Return the JSON response content as a dictionary
        return APIResponse(**response.json())

    def update(
        self,
        voice_id: Optional[str] = None,
        voice_name: Optional[str] = None,
        new_voice_file_path: Optional[str] = None,
        new_voice_name: str = '',
        new_voice_tags: Optional[List[str]] = None,
    ) -> APIResponse[dict]:
        """
        Update a voice by its ID or name.

        Parameters
        ----------
        voice_id : Optional[str]
            The ID of the voice to update.
        voice_name : Optional[str]
            The name of the voice to update. This does not need to be provided if voice_id
            is provided.
        new_voice_file_path: Optional[str]
            The file path for the new audio file to use for the cloned voice.
        new_voice_name: str
            The new name updated name for the voice.
        new_voice_tags: Optional[List[str]]
            The new tags for the voice.

        Returns
        -------
        APIResponse[dict]
            response.data will contain a success message with the updated fields of the voice.

        Raises
        ------
        httpx.HTTPStatusError
            If the request to update the voice fails. This will usually trigger if you do not have
            permissions to update the voice.
        """

        # Accept case if user only provide name
        if not voice_id:
            try:
                voice_id = self._get_voice_id_from_name(voice_name=voice_name)
            except ValueError as e:
                raise ValueError(
                    f'No voice found with the name {voice_name}. You cannot update this voice.'
                )

        # Convert voice tags to a string
        if new_voice_tags:
            new_voice_tags = (', ').join(new_voice_tags)

        # If voice_file is not given
        if new_voice_file_path is None:
            params = {
                'new_voice_tags': new_voice_tags,
                'new_voice_file': new_voice_file_path,
            }
            files = {}

        # If voice file is given
        else:
            params = {
                'new_voice_tags': new_voice_tags,
            }
            files = {'new_voice_file': open(new_voice_file_path, 'rb')}

        # Call API
        response = httpx.patch(
            f'{self.http_url}/voices/{voice_id}?new_voice_name={new_voice_name}',
            params=params,
            headers=self.headers,
            timeout=self.timeout,
            files=files,
        )

        # Handle response errors
        if not response.is_success:
            raise httpx.HTTPStatusError(
                f'Failed to update voice. Status code: {response.status_code}. Error: {response.text}',
                request=response.request,
                response=response,
            )

        # Return the JSON response content as a dictionary
        return APIResponse(**response.json())

    def delete(self, voice_id: str = None, voice_name=None) -> APIResponse[dict]:
        """
        Delete a voice by its ID.

        Parameters
        ----------
        voice_id : str
            The ID of the voice to be deleted.

        Returns
        -------
        APIResponse[dict]
            response.data will contain a success message with the updated fields of the voice.

        Raises
        ------
        httpx.HTTPStatusError
            If the request to delete the voice fails. This will usually trigger if you do not have
            permissions to delete the voice.
        """
        if not voice_id:
            try:
                voice_id = self._get_voice_id_from_name(voice_name=voice_name)
            except ValueError as e:
                raise ValueError(
                    f'No voice found with the name {voice_name}. You cannot Delete this voice.'
                )

        response = httpx.delete(
            f'{self.http_url}/voices/{voice_id}',
            headers=self.headers,
            timeout=self.timeout,
        )

        # Handle response errors
        if not response.is_success:
            raise httpx.HTTPStatusError(
                f'Failed to delete voice. Status code: {response.status_code}. Error: {response.text}',
                request=response.request,
                response=response,
            )

        # Return the JSON response content as a dictionary
        return APIResponse(**response.json())
